﻿using UnityEngine;
using System.Collections;

public class Target : MonoBehaviour {

	public float m_size = 1.0f;
	public float m_minYOffset = 0.1f;
	public float m_holdTime = 1.0f;
	public float m_jiggleDown = 0.1f;
	public float m_jiggleUp = 0.05f;
	public float m_vacuumJiggle = 0.1f;
	public float m_vacuumMult = 10.0f;
	public int m_rescueScore = 5;
	public Camera m_camera;
	public Transform m_transform;
	public Rigidbody2D m_rigidbody;
	public AudioSource m_grabSound;
	public AudioSource m_pointSound;
	public AudioSource m_dyingSound;
	public Score m_score;
	public Lives m_lives;
	public Transform m_spriteT;
	public Transform m_vacuumEndT;
	public AttachedExplosion m_explosion;

	private Vector3 m_startPos;
	private float m_sleepTimer = 0.0f;
	private State m_state = State.Finished;
	private float m_holdTimer = 0.0f;
	private bool m_waveEnding = false;
	private Wave.Target m_data;
	private Vector3 m_startScale = Vector3.one;

	private enum State
	{
		Sleeping,
		Running,
		Holding,
		Finished
	}

	void Awake()
	{
		m_startScale = m_spriteT.localScale;
	}

	void Start()
	{
		Vector3 m_maxPos = m_camera.ScreenToWorldPoint (new Vector3 (Screen.width, Screen.height));
		m_maxPos.z = 0.0f;
		m_maxPos += new Vector3 (-m_size, m_size, 0.0f);
		m_transform.position = m_maxPos;
		m_spriteT.localScale = m_startScale;
		m_rigidbody.gravityScale = 0.0f;
		m_rigidbody.velocity = Vector2.zero;
		m_rigidbody.angularVelocity = 0.0f;
	}

	void Update ()
	{
		if(State.Holding == m_state)
		{
			m_holdTimer -= Time.deltaTime;

			if(0 >= m_holdTimer)
			{
				m_score.AddScore(m_rescueScore);
				if(m_waveEnding)
				{
					Finish();
				}
				else
				{
					ResetPos();
				}
				m_pointSound.Play();
			}
		}
		else if(State.Sleeping == m_state)
		{
			// No gravity if asleep.
			m_sleepTimer -= Time.deltaTime;
			if(0 >= m_sleepTimer)
			{
				m_rigidbody.gravityScale = Random.Range(m_data.m_minAcc, m_data.m_maxAcc);
				m_rigidbody.drag = Random.Range(m_data.m_minDrag, m_data.m_maxDrag);
				m_state = State.Running;
			}
		}

		if(State.Holding == m_state || State.Running == m_state)
		{
			Vector3 pos = m_transform.position;
			Vector3 minPos = m_camera.ScreenToWorldPoint (new Vector3(0.0f, m_minYOffset*Screen.height, 0.0f));
			if(pos.y < minPos.y - m_size)
			{
				m_lives.SubtractLife();

				if(m_waveEnding)
				{
					Finish();
				}
				else
				{
					ResetPos ();
				}

				m_dyingSound.Play();
				m_explosion.Fire (pos);
			}
		}
	}

	void FixedUpdate()
	{
		if(State.Holding == m_state)
		{
			// Suck towards the end of the arm
			Vector2 toVacuum = m_vacuumEndT.position - transform.position;
			toVacuum *= m_vacuumMult;
			toVacuum.x *= Random.Range(m_vacuumJiggle, 2.0f - m_vacuumJiggle);
			toVacuum.y *= Random.Range(m_vacuumJiggle, 2.0f - m_vacuumJiggle);
			m_rigidbody.AddForce(toVacuum);

			// Jiggle the scale
			// localScale: startScale -> 0
			// jingle: -m_jiggleDown -> m_jiggleUp
			float jiggleDown = Mathf.Lerp(m_jiggleUp, -m_jiggleDown, (m_spriteT.localScale.x / m_startScale.x));
			float size = Random.Range(jiggleDown, m_jiggleUp);
			m_spriteT.localScale += new Vector3(size, size, 0.0f);
		}
	}

	void OnTriggerEnter2D(Collider2D other)
	{
		if(State.Running == m_state)
		{
			m_state = State.Holding;
			m_holdTimer = m_holdTime;
			m_grabSound.Play();
		}
	}

	void OnTriggerExit2D(Collider2D other)
	{
		if(State.Holding == m_state)
		{
			m_state = State.Running;
			m_spriteT.localScale = m_startScale;
		}
	}

	void ResetPos ()
	{
		Vector3 m_minPos = m_camera.ScreenToWorldPoint (new Vector3 (0.0f, 0.0f));
		m_minPos += new Vector3 (m_size, -m_size, 0.0f);
		Vector3 m_maxPos = m_camera.ScreenToWorldPoint (new Vector3 (Screen.width, Screen.height));
		m_maxPos += new Vector3 (-m_size, m_size, 0.0f);

		m_state = State.Sleeping;
		m_sleepTimer = Random.Range (m_data.m_minWait, m_data.m_maxWait);
		m_transform.position = new Vector3(Random.Range(m_minPos.x, m_maxPos.x), m_maxPos.y, 0.0f);
		m_spriteT.localScale = m_startScale;
		m_transform.Rotate (0.0f, 0.0f, Random.Range(0, 360.0f));
		m_rigidbody.gravityScale = 0.0f;
		m_rigidbody.velocity = Vector2.zero;
		m_rigidbody.angularVelocity = Random.Range (-m_data.m_maxAngularVel, m_data.m_maxAngularVel);
	}
	
	void WaveEnding()
	{
		m_waveEnding = true;
	}

	public bool IsFinished { get { return State.Finished == m_state; } }

	public void Reset(Wave.Target data)
	{
		m_data = data;
		m_waveEnding = false;
		m_explosion.Hide ();
		ResetPos ();
	}

	public void Finish()
	{
		Vector3 m_maxPos = m_camera.ScreenToWorldPoint (new Vector3 (Screen.width, Screen.height));
		m_maxPos.z = 0;
		m_maxPos += new Vector3 (-m_size, m_size, 0.0f);
		m_transform.position = m_maxPos;
		m_spriteT.localScale = m_startScale;
		m_rigidbody.velocity = Vector2.zero;
		m_rigidbody.angularVelocity = 0.0f;
		m_rigidbody.gravityScale = 0.0f;
		m_state = State.Finished;
	}
}
